//! Platform-specific extensions for Unix platforms.
//!
//! This corresponds to [`async_std::os::unix`].
//!
//! [`async_std::os::unix`]: https://docs.rs/async-std/latest/async_std/os/unix/

pub mod net;
